package chaincode

import (
	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"encoding/json"
	"github.com/pkg/errors"
	"reflect"
	"gitlab.smartblocklab.com/smartcontracts/common/security/actions"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
)

// Пример наследования данной структуры
//type AccountManagerImpl struct {
//	*manager.ManagerImpl
//}

func GetManager(adapter StubInterfaceAdapter, timeGetter TimeGetter) IManager {
	GlobalLogger.Infof("GetManager")
	manager := &ManagerImpl{adapter, timeGetter}
	return manager
}

type ManagerImpl struct {
	Adapter    StubInterfaceAdapter
	TimeGetter TimeGetter
}

func (this ManagerImpl) GetTime() int64 {
	GlobalLogger.Infof("managerImpl.GetTime")
	tt, err := this.TimeGetter()
	if err != nil {
		GlobalLogger.Error(err.Error())
		return 0
	} else {
		return tt
	}
}

func (this ManagerImpl) SetAdapter(new StubInterfaceAdapter) {
	GlobalLogger.Infof("managerImpl.SetAdapter")
	this.Adapter = new
}

func (this ManagerImpl) SetTimeGetter(new TimeGetter) {
	GlobalLogger.Infof("managerImpl.SetTimeGetter")
	this.TimeGetter = new
}

func (this ManagerImpl) GetAdapter() StubInterfaceAdapter {
	GlobalLogger.Infof("managerImpl.GetAdapter")
	return this.Adapter
}

func (this ManagerImpl) GetTimeGetter() TimeGetter {
	GlobalLogger.Infof("managerImpl.GetTimeGetter")
	return this.TimeGetter
}

func (this ManagerImpl) GetState(key string) ([]byte, error) {
	GlobalLogger.Infof("managerImpl.GetState")
	return this.Adapter.GetState(key)
}

func (this ManagerImpl) PutState(key string, value []byte) error {
	GlobalLogger.Infof("managerImpl.PutState by key %s", key)
	return this.Adapter.PutState(key, value)
}

func (this ManagerImpl) DelState(key string) error {
	GlobalLogger.Infof("managerImpl.DelState %s", key)
	return this.Adapter.DelState(key)
}

// TODO REMOVE >>> DEPRECATED <<<
func (this ManagerImpl) GetQueryResult(query string) (shim.StateQueryIteratorInterface, error) {
	GlobalLogger.Infof("managerImpl.GetQueryResult %s", query)
	return this.Adapter.GetQueryResult(query)
}

func (this ManagerImpl) InvokeChaincode(chaincodeName string, args [][]byte, channel string) pb.Response {
	GlobalLogger.Infof("managerImpl.InvokeChaincode")
	return this.Adapter.InvokeChaincode(chaincodeName, args, channel)
}

func (this ManagerImpl) GetTxID() string {
	GlobalLogger.Infof("managerImpl.GetTxID")
	return this.Adapter.GetTxID()
}

func (this ManagerImpl) Save(obj ValidObject) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.Save")
	if obj == nil {
		return nil, errors.New("can't save nil object")
	}
	if reflect.ValueOf(obj).Kind() != reflect.Ptr {
		return nil, errors.New("check newObject is not reference")
	}

	err := obj.Validate()
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	objByteArray, err := json.Marshal(obj)
	if err != nil {
		return nil, GlobalLogger.Errorf("failed to Marshal. error %s", err)
	}

	key, err := GetKeyByArray(obj.GetKey(this))
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	err = this.PutState(key, objByteArray)
	if err != nil {
		return nil, GlobalLogger.Errorf("Failed to PutState. Error: %s", err)
	}

	return obj, nil
}

func (this ManagerImpl) IsExists(obj ValidObject) (bool, error) {
	GlobalLogger.Infof("managerImpl.IsExists")
	key, err := GetKeyByArray(obj.GetKey(this))
	if err != nil {
		return false, GlobalLogger.Error(err.Error())
	}

	bytes, err := this.GetState(key)
	if err != nil {
		return false, GlobalLogger.Error(err.Error())
	}

	if bytes == nil || len(bytes) == 0 {
		return false, nil
	}

	err = json.Unmarshal(bytes, &obj)
	if err != nil {
		return false, GlobalLogger.Error(err.Error())
	}

	return true, nil
}

func (this ManagerImpl) GetIfExists(newObject ValidObject) (bool, error) {
	GlobalLogger.Infof("managerImpl.GetIfExists")
	key, err := GetKeyByArray(newObject.GetKey(this))
	if err != nil {
		return false, GlobalLogger.Error(err.Error())
	}

	bytes, err := this.GetObject(newObject.New(), key)
	if err != nil {
		return false, GlobalLogger.Error(err.Error())
	}

	return bytes != nil, nil
}

func (this ManagerImpl) AddObject(newObject ValidObject) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.AddObject")
	if newObject == nil {
		return nil, errors.New("newObject can't be nil")
	}
	if reflect.ValueOf(newObject).Kind() != reflect.Ptr {
		return nil, errors.New("check newObject is reference")
	}

	err := newObject.Validate()
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	isExists, err := this.IsExists(newObject)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	if isExists {
		return nil, errors.New("ALREADY_EXISTS")
	}

	err = newObject.CheckBeforeAdd(this)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	return this.Save(newObject)
}

func (this ManagerImpl) UpdateObject(newObject ValidObject) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.UpdateObject")
	if newObject == nil {
		return nil, errors.New("newObject can't be nil")
	}
	if reflect.ValueOf(newObject).Kind() != reflect.Ptr {
		return nil, errors.New("check newObject is reference")
	}

	err := newObject.Validate()
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	isExists, err := this.GetIfExists(newObject)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	if !isExists {
		return nil, errors.New("NOT FOUND UPDATED OBJECT BY KEY")
	}

	err = newObject.CheckBeforeUpdate(newObject.New(), this)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	return this.Save(newObject)
}

func (this ManagerImpl) GetObject(expectedObject ValidObject, keys ...string) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.GetObject")
	return this.GetObjectByKeys(expectedObject, keys)
}

func (this ManagerImpl) GetObjectByKeys(expectedObject ValidObject, keys []string) (ValidObject, error) {
	GlobalLogger.Infof("managerImpl.GetObjectByKeys")
	key, err := GetKeyByArray(keys)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}

	bytes, err := this.GetState(key)
	if err != nil {
		return nil, GlobalLogger.Errorf("GetState by key %s. Error: %s", key, err)
	}

	if bytes == nil || len(bytes) == 0 {
		//return nil, GlobalLogger.Error("NOT_FOUND")
		return nil, nil
	}

	err = json.Unmarshal(bytes, expectedObject)
	if err != nil {
		return nil, GlobalLogger.Error(err.Error())
	}
	return expectedObject, nil
}

func (this ManagerImpl) PrepareInvokeChaincodeParam(channelName string, chaincodeName string, functionName string, arg interface{}) ([][]byte, error) {
	preference, err := actions.FindSecurityPreferenceByFunctionName(chaincodeName, functionName)
	if err != nil {
		return nil, err
	}

	signedRequest := requests.CCFullSignedRequest{
		UserSignature:   requests.UserSignature{},
		SystemSignature: requests.UserSignature{},
		Parameters: requests.CCParameters{
			UserParams: requests.CCParameter{},
			SystemParams: requests.CCParameter{
				UserID:   "NOT_APPLICABLE",
				ActionId: preference.GetActionID(channelName),
				Param:    arg,
			},
		},
	}

	param, err := json.Marshal(signedRequest)
	if err != nil {
		return nil, err
	}

	return [][]byte{[]byte(preference.FunctionName), param}, nil
}

func (this ManagerImpl) SignedInvokeChaincode(channelName string, chaincodeName string, functionName string, arg interface{}) pb.Response {
	GlobalLogger.Infof("managerImpl.SignedInvokeChaincode")

	bytes, err := this.PrepareInvokeChaincodeParam(channelName, chaincodeName, functionName, arg)
	if err != nil {
		return shim.Error(err.Error())
	}

	return this.InvokeChaincode(chaincodeName, bytes, channelName)
}

func (this ManagerImpl) InvokeChaincode_GetUser(UserID string) pb.Response {
	GlobalLogger.Infof("managerImpl.InvokeChaincode_GetUser")
	return this.SignedInvokeChaincode(invoke.UserChannel, invoke.UserCC, invoke.GetUserFunction, requests.Id{Id: UserID})
}

func (this ManagerImpl) InvokeChaincode_VerifyUserSignature(arg interface{}) pb.Response {
	GlobalLogger.Infof("managerImpl.InvokeChaincode_VerifyUserSignature")
	return this.SignedInvokeChaincode(invoke.UserChannel, invoke.UserCC, invoke.VerifyUserSignatureFunction, arg)
}

func GetKeyByArray(keys []string) (string, error) {
	if len(keys) == 0 {
		return "", GlobalLogger.Error("keys list is empty")
	} else if len(keys) == 1 && len(keys[0]) == 0 {
		return "", GlobalLogger.Error("key is empty")
	}

	total := ""
	for _, num := range keys {
		if len(num) == 0 {
			return "", GlobalLogger.Errorf("key in array %s of key can't empty", keys)
		}

		total += num + "."
	}

	sz := len(total)
	if sz > 0 && total[sz-1] == '.' {
		total = total[:sz-1]
	}
	return total, nil
}
