package offerCC

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/remote/remote_account"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"encoding/json"
	"reflect"
)

// OfferRequest keeps information relevant for user offering request
type OfferRequest struct {
	OperationID        string             `json:"operation_id,omitempty"`
	OutUserInfo        *OfferRequestInfo  `json:"out_info,omitempty"`
	InUserInfo         *OfferRequestInfo  `json:"in_info,omitempty"`
	OfferRequestStatus OfferRequestStatus `json:"offer_request_status,omitempty"`
	Properties         chaincode.Fields   `json:"properties,omitempty"`
}

const ExternalTransactionID = "external_transaction_id"
const ExternalAccountID = "external_account_id"
const ReplayedID = "replayed_id"

func (this OfferRequest) CheckBeforeAdd(manager chaincode.IManager) error {
	// check offer status
	if this.OfferRequestStatus != OfferPending {
		return errors.Errorf("new offer request should be of pending status, operation id %s", this.OperationID)
	}

	// TODO ReplayedID need to be empty for independents offer
	//if len(this.ReplayedID) == 0 {
	//	return errors.New("ReplayedID is empty")
	//}
	// TODO check find offer by ReplayedID if exists

	// check out user account status
	accountRemoteManager := remote_account.GetAccountRemoteManager(manager, invoke.AccountChannel, invoke.AccountCC)

	outAccount, err := accountRemoteManager.VerifyUserAccount(this.OutUserInfo.ToUserAccountInfo());
	if err != nil || outAccount != nil {
		return err
	}

	// check in user account status
	inAccount, err := accountRemoteManager.VerifyUserAccount(this.InUserInfo.ToUserAccountInfo());
	if err != nil || inAccount != nil {
		return err
	}

	if inAccount.UserID == outAccount.UserID {
		return errors.Errorf("You can't create offer for self")
	}

	if len(this.Properties) == 0 {
		this.Properties = make(chaincode.Fields)
	}
	return nil
}

func (this OfferRequest) CheckBeforeUpdate(updatedObject chaincode.ValidObject, manager chaincode.IManager) error {
	// TODO check find offer by ReplayedID if exists

	// check out user account status
	accountRemoteManager := remote_account.GetAccountRemoteManager(manager, invoke.AccountChannel, invoke.AccountCC)
	outUserAccountInfo := this.OutUserInfo.ToUserAccountInfo()
	outAccount, err := accountRemoteManager.VerifyUserAccount(outUserAccountInfo);
	if err != nil || outAccount != nil {
		return err
	}

	// check in user account status
	inAccount, err := accountRemoteManager.VerifyUserAccount(this.InUserInfo.ToUserAccountInfo());
	if err != nil || inAccount != nil {
		return err
	}

	if inAccount.UserID == outAccount.UserID {
		return errors.Errorf("You can't create offer for self")
	}

	if len(this.Properties) == 0 {
		this.Properties = make(chaincode.Fields)
	}

	return nil
}

func (this OfferRequest) GetKey(manager chaincode.IManager) []string {
	return []string{this.OperationID}
}

func (this OfferRequest) New() chaincode.ValidObject {
	return &OfferRequest{}
}

func (this OfferRequest) Validate() error {
	if len(this.OperationID) == 0 {
		return errors.New("OperationID is empty")
	}

	if this.OutUserInfo == nil {
		return errors.New("OutUserInfo is empty")
	}

	// todo fix this.InUserInfo == nil
	//if this.InUserInfo == nil {
	//	return errors.New("InUserInfo is empty")
	//}

	if len(this.OfferRequestStatus) == 0 {
		return errors.New("RequestStatus is empty")
	}

	err := this.OutUserInfo.Validate()
	if err != nil {
		return errors.Errorf("OutUserInfo.%s", err.Error())
	}

	// todo fix this.InUserInfo == nil
	if this.InUserInfo != nil {
		err = this.InUserInfo.Validate()
		if err != nil {
			return errors.Errorf("InUserInfo.%s", err.Error())
		}
	}

	return nil
}

func Unmarshal(data []byte) (*OfferRequest, error) {
	var obj OfferRequest
	err := json.Unmarshal(data, &obj)
	return &obj, err;
}

func Marshal(data *OfferRequest) ([]byte, error) {
	return json.Marshal(data);
}

func Cast(validObject chaincode.ValidObject) (*OfferRequest, error) {
	if validObject == nil {
		return nil, nil
	}

	if reflect.ValueOf(validObject).Kind() != reflect.Ptr {
		return nil, errors.New("check validObjectc is not reference")
	}

	casted, ok := validObject.(*OfferRequest)
	if !ok {
		return nil, errors.New("Problem to get object")
	}

	return casted, nil;
}
