package offerCC

import (
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
	"encoding/json"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
)

// toUserAccountInfo converts OfferRequestInfo into UserAccountInfo
// representation
func (o OfferRequestInfo) ToUserAccountInfo() *accountCC.UserAccountInfo {
	return &accountCC.UserAccountInfo{
		UserID: o.UserID,
		TxnAccountInfo: &accountCC.TxnAccountInfo{
			WalletID:  o.WalletID,
			AccountID: o.AccountID,
		},
	}
}

// UsersOffers encapsulates all users offers
type UsersOffers struct {
	UserID string          `json:"user_id,omitempty"`
	Offers []*OfferRequest `json:"user_offers,omitempty"`
}

// OffersList keeps list of offers
type OffersList struct {
	Offers []*OfferRequest `json:"user_offers,omitempty"`
}

// OfferRequestInfo capture offer information relevant to request
type OfferRequestInfo struct {
	*accountCC.UserAccountInfo
	Amount nums.PlainCurrencyNumber `json:"amount,omitempty"`
}

func (offerRequestInfo OfferRequestInfo) Validate() error {
	if offerRequestInfo.UserAccountInfo == nil {
		return errors.New("UserAccountInfo is empty")
	}

	err := offerRequestInfo.UserAccountInfo.Validate()
	if err != nil {
		return err
	}

	err = offerRequestInfo.Amount.Validate()
	if err != nil {
		return err
	}
	return nil
}

// OfferRequestStatus type to represent offer request status
type OfferRequestStatus string

// Enumerate possible statuses for users offers requests
const (
	OfferPending   OfferRequestStatus = "OfferPending"
	OfferExecuting OfferRequestStatus = "OfferExecuting"
	OfferAccepted  OfferRequestStatus = "OfferAccepted"
	OfferRejected  OfferRequestStatus = "OfferRejected"
	OfferCanceled  OfferRequestStatus = "OfferCanceled"
	OfferCompleted OfferRequestStatus = "OfferCompleted"
)

func CollectOffers(query string, o *UsersOffers, adapter chaincode.StubInterfaceAdapter) (*UsersOffers, error) {
	it, err := adapter.GetQueryResult(query)
	if err != nil {
		return nil, errors.Errorf("failed to retreive all outcoming user request for user id %s, due to %s", o.UserID, err)
	}

	for it.HasNext() {
		kv, err := it.Next()
		if err != nil {
			return nil, errors.Errorf("failed to get next outcoming user request for user id %s, due to %s", o.UserID, err)
		}
		var offer OfferRequest
		err = json.Unmarshal(kv.Value, &offer)
		if err != nil {
			return nil, errors.Errorf("failed to deserialize outcoming user request for user id %s, due to %s", o.UserID, err)
		}

		o.Offers = append(o.Offers, &offer)
	}

	return o, nil
}

type GetUserIncomingRequestsByStatusArg struct {
	UserID             string `json:"user_id"`
	OfferRequestStatus string `json:"offer_request_status"`
}

type GetRequestsByStatusArg struct {
	OfferRequestStatus string `json:"offer_request_status"`
}

type GetUserOutgoingRequestsByStatusArg struct {
	UserID             string `json:"user_id"`
	OfferRequestStatus string `json:"offer_request_status"`
}

type GetHistoryUserOutgoingRequestsArg struct {
	UserID string `json:"user_id"`
}

type GetHistoryUserIncomingRequestsArg struct {
	UserID string `json:"user_id"`
}

type SetOfferStatusByOperationIdArg struct {
	OperationID        string             `json:"operation_id"`
	OfferRequestStatus OfferRequestStatus `json:"offer_request_status"`
}

type PutOfferPropertiesByOperationIDArg struct {
	OperationID string `json:"operation_id"`
	Key         string `json:"key,omitempty"`
	Value       string `json:"value,omitempty"`
}

type RequestsByStatusAndCurrencyType struct {
	OfferRequestStatus OfferRequestStatus     `json:"offer_request_status,omitempty"`
	OutUserInfo        RequestsByCurrencyType `json:"out_info,omitempty"`
}

type PlainCurrencyNumberVar struct {
	CurrencyType string `json:"currency_type"`
}

type RequestsByCurrencyType struct {
	Amount PlainCurrencyNumberVar `json:"amount,omitempty"`
}

type GetRequestsByStatusAndCurrencyTypeArg struct {
	OfferRequestStatus string `json:"offer_request_status"`
	CurrencyType       string `json:"currency_type"`
}
