package userCC

import (
	"encoding/pem"
	"crypto/x509"
	"crypto/ecdsa"
	"crypto/sha256"
	"math/big"
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/model/system_user"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
)

func VerifyPublicKey(manager chaincode.IManager, sourceUser User) error {
	if sourceUser.CustomerType == Root {
		return nil
	}

	if sourceUser.CustomerType == Company {
		rootUser := &User{}
		object, err := manager.GetObject(rootUser, system_user.ROOT_USER_ID);
		if err != nil || object == nil {
			return errors.Errorf("failed to get RootUser, due to %s", err)
		}

		for _, userPublicKey := range sourceUser.PublicKeys {
			return verifyPublicKeyCompany(userPublicKey, sourceUser.CustomerType, rootUser)
		}

	} else {
		for _, userPublicKey := range sourceUser.PublicKeys {
			_, err := verifyPublicKey(userPublicKey)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// verifyPublicKey function to verify whenever user public key indeed belongs
// to the user, by validating user signature on the public key
func verifyPublicKey(userPublicKey UserPublicKey) ([]byte, error) {
	if userPublicKey.Revoked {
		return nil, errors.Errorf("PublicKey with alias %s was marked as revoked", userPublicKey.Alias)
	}

	// get public key from user public key
	block, _ := pem.Decode([]byte(userPublicKey.PublicKey))
	if block == nil {
		return nil, errors.New("failed to parse PEM block containing the public key")
	}

	basePubKey, err := x509.ParsePKIXPublicKey(block.Bytes)
	if err != nil {
		return nil, errors.Errorf("failed to parse DER encoded public key: " + err.Error())
	}

	var pubKey *ecdsa.PublicKey

	switch pub := basePubKey.(type) {
	case *ecdsa.PublicKey:
		pubKey = pub
	default:
		return nil, errors.Errorf("Unknown type of public key")
	}

	// hash user public key
	hasher := sha256.New()
	hasher.Write([]byte(userPublicKey.PublicKey))
	digest := hasher.Sum(nil)

	R := &big.Int{}
	R.SetString(userPublicKey.OwnerSignature.R, 16)
	S := &big.Int{}
	S.SetString(userPublicKey.OwnerSignature.S, 16)

	// verify signature on the signature based on user public key, declared
	// and extracted from the public key
	if !ecdsa.Verify(pubKey, digest, R, S) {
		return nil, errors.Errorf("failed to verify user public key for alias %s", userPublicKey.Alias)
	}

	return digest, nil
}

func verifyPublicKeyCompany(userPublicKey UserPublicKey, userType CustomerType, rootUser *User) error {
	if userType != Company {
		return errors.Errorf("invalid user type '%s' to call this function need have Company user type ", userType)
	}

	digest, err := verifyPublicKey(userPublicKey)
	if err != nil {
		return err
	}

	rootR := &big.Int{}
	rootR.SetString(userPublicKey.AuthoritySignature.R, 16)
	rootS := &big.Int{}
	rootS.SetString(userPublicKey.AuthoritySignature.S, 16)

	// get public key from user public key
	if len(rootUser.PublicKeys) == 0 {
		return errors.Errorf("RootUser has no any keys")
	}

	// у rootUser есть только один PublicKey
	for _, publicKey := range rootUser.PublicKeys {
		block, _ := pem.Decode([]byte(publicKey.PublicKey))
		if block == nil {
			return errors.New("failed to parse PEM block containing the public key")
		}

		basePubKey, err := x509.ParsePKIXPublicKey(block.Bytes)
		if err != nil {
			return errors.Errorf("failed to parse DER encoded public key: " + err.Error())
		}

		var rootPubKey *ecdsa.PublicKey

		switch pub := basePubKey.(type) {
		case *ecdsa.PublicKey:
			rootPubKey = pub
		default:
			return errors.Errorf("Unknown type of public key")
		}

		// verify signature on the signature based on user public key, declared
		// and extracted from the public key
		if !ecdsa.Verify(rootPubKey, digest, rootR, rootS) {
			return errors.Errorf("failed to verify user public key for alias %s with RootUser", userPublicKey.Alias)
		}
		// т.к у rootUser есть только один PublicKey то выходим из программы после проверки
		return nil
	}

	return nil
}
